import logging
from uuid import UUID

from django.db import connections

logger = logging.getLogger(__name__)


def delete_element(element_id: UUID) -> None:
    assert isinstance(element_id, UUID)
    with connections['default'].cursor() as cursor:
        # Remove transcription-entity links
        cursor.execute("""
        DELETE FROM documents_transcriptionentity WHERE transcription_id IN (
            SELECT t.id FROM documents_transcription t
            LEFT JOIN documents_elementpath elementpath USING (element_id)
            WHERE t.element_id = %(id)s OR elementpath.path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} transcription-entity links")

        # Remove transcriptions
        cursor.execute("""
        DELETE FROM documents_transcription
        WHERE element_id = %(id)s
        OR element_id IN (
            SELECT element_id FROM documents_elementpath WHERE path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} transcriptions")

        # Remove classifications
        cursor.execute("""
        DELETE FROM documents_classification
        WHERE element_id = %(id)s
        OR element_id IN (
            SELECT element_id FROM documents_elementpath WHERE path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} classifications")

        # Remove element metadata
        cursor.execute("""
        DELETE FROM documents_metadata
        WHERE element_id = %(id)s
        OR element_id IN (
            SELECT element_id FROM documents_elementpath WHERE path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} metadata")

        # Remove element-process links
        cursor.execute("""
        DELETE FROM process_processelement
        WHERE element_id = %(id)s
        OR element_id IN (
            SELECT element_id FROM documents_elementpath WHERE path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} usage from process as selection")

        # Set elements directly on processes to None
        cursor.execute("""
        UPDATE process_process
        SET element_id = NULL
        WHERE element_id = %(id)s
        OR element_id IN (
            SELECT element_id FROM documents_elementpath WHERE path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} usage from process as element")

        # Set folders references on training processes to None
        cursor.execute("""
        UPDATE process_process
        SET train_folder_id = NULL
        WHERE train_folder_id = %(id)s
        OR train_folder_id IN (
            SELECT element_id FROM documents_elementpath WHERE path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} usage from process as train folder")
        cursor.execute("""
        UPDATE process_process
        SET validation_folder_id = NULL
        WHERE validation_folder_id = %(id)s
        OR validation_folder_id IN (
            SELECT element_id FROM documents_elementpath WHERE path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} usage from process as validation folder")
        cursor.execute("""
        UPDATE process_process
        SET test_folder_id = NULL
        WHERE test_folder_id = %(id)s
        OR test_folder_id IN (
            SELECT element_id FROM documents_elementpath WHERE path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} usage from process as test folder")

        # Remove user selections
        cursor.execute("""
        DELETE FROM documents_selection selection
        WHERE element_id = %(id)s
        OR element_id IN (
            SELECT element_id FROM documents_elementpath WHERE path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} user selections")

        # Remove workers activity on this element
        cursor.execute("""
        DELETE FROM process_workeractivity
        WHERE element_id = %(id)s
        OR element_id IN (
            SELECT element_id FROM documents_elementpath WHERE path && ARRAY[%(id)s]
        )
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} worker activities")

        # Remove element paths and elements simultaneously
        cursor.execute("""
        WITH children_ids (id) AS (
            DELETE FROM documents_elementpath
            WHERE element_id = %(id)s OR path && ARRAY[%(id)s]
            RETURNING element_id
        )
        DELETE FROM documents_element element
        USING children_ids
        WHERE element.id = children_ids.id
        """, {'id': element_id})
        logger.info(f"Deleted {cursor.rowcount} child elements with paths")
