#!/usr/bin/env python3
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction

from arkindex.documents.models import ElementType
from arkindex.project.argparse import CorpusArgument


class Command(BaseCommand):
    help = 'Merge an element type with another.'

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '-c',
            '--corpus',
            help='ID of the corpus to find types on.',
            type=CorpusArgument(),
            required=True,
        )
        parser.add_argument(
            '-s',
            '--source',
            help='Slug of the type that will be merged and removed.',
            required=True,
        )
        parser.add_argument(
            '-d',
            '--destination',
            help="Slug of the type that will receive the other type's elements.",
            required=True,
        )

    @transaction.atomic
    def handle(self, corpus, source, destination, verbosity=0, **options):
        try:
            source_type = corpus.types.get(slug=source)
        except ElementType.DoesNotExist:
            raise CommandError(f'Source type {source} not found in {corpus}')

        try:
            destination_type = corpus.types.get(slug=destination)
        except ElementType.DoesNotExist:
            raise CommandError(f'Destination type {destination} not found in {corpus}')

        if source_type == destination_type:
            raise CommandError('Source and destination types cannot be the same.')

        affected = corpus.elements.filter(type=source_type).count()

        self.stdout.write(f'Updating {affected} {source} elements to {destination}…')
        corpus.elements.filter(type=source_type).update(type=destination_type)

        # Update the corpus top level type if required
        if corpus.top_level_type_id == source_type.id:
            self.stdout.write('Updating top level type of this corpus to the destination type.')
            corpus.top_level_type_id = destination_type.id
            corpus.save()

        self.stdout.write(f'Removing type {source}…')
        source_type.delete()

        self.stdout.write(self.style.SUCCESS(f'Successfully merged type {source} into {destination}.'))
        self.stdout.write('Committing transaction…')
