#!/usr/bin/env python3
from typing import NamedTuple

from django.core.management.base import BaseCommand, CommandError

from arkindex.documents.models import ElementType
from arkindex.project.argparse import CorpusArgument

# x and y of top left and bottom right points
BBox = NamedTuple(
    "BBox", [("left", int), ("top", int), ("right", int), ("bottom", int)]
)


def compute_polygon_area(polygon: BBox):
    return (polygon.bottom - polygon.top) * (polygon.right - polygon.left)


def reorder_polygon(polygon):
    """Reorders the points composing the polygon and convert it to a BBox.
    This transforms this polygon into a bounding box, a rectangle, to make later computations
    like area easier.
    """
    heights = {h for _, h in polygon}
    widths = {w for w, _ in polygon}
    return BBox(
        left=min(widths),
        top=min(heights),
        right=max(widths),
        bottom=max(heights),
    )


def compute_area_overlap(parent_polygon: BBox, child_polygon: BBox):
    """Compute intersection pourcentage of child_polygon in parent polygon
    """
    left = max(parent_polygon.left, child_polygon.left)
    top = max(parent_polygon.top, child_polygon.top)
    right = min(parent_polygon.right, child_polygon.right)
    bottom = min(parent_polygon.bottom, child_polygon.bottom)
    height = bottom - top
    width = right - left
    if min(height, width) <= 0:
        # otherwise might get reversed rectangles that are bigger than cells
        return 0
    return height * width


def is_overlapping_enough(parent_polygon: BBox, child_polygon: list, threshold: float):
    """
    Checks if the intersection between the child's polygon and the parent's polygon is higher than the threshold.
    """
    # Reorder the child's polygon and convert it to a BBox
    polygon = reorder_polygon(child_polygon)
    child_area = compute_polygon_area(polygon)

    # Early return if polygon is too small to compute an area
    # Prevent division by zero if any should occur
    if child_area == 0:
        return False

    # Check the intersection value
    area_overlap = compute_area_overlap(parent_polygon=parent_polygon,
                                        child_polygon=polygon)

    return (area_overlap / child_area) > threshold


class Command(BaseCommand):
    help = 'Move elements to closest geographical parent.'

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '-c',
            '--corpus',
            help='ID of the corpus to find types on.',
            type=CorpusArgument(),
            required=True,
        )
        parser.add_argument(
            '-s',
            '--single-page-type',
            help='Slug of the type corresponding to a Single Page.',
            required=False,
            default="single_page"
        )
        parser.add_argument(
            '-d',
            '--double-page-type',
            help="Slug of the type corresponding to a Double Page.",
            required=False,
            default="double_page"
        )
        parser.add_argument(
            '--element',
            help="Arkindex element id to process",
        )

    def handle(self, corpus, single_page_type, double_page_type, element=None, verbosity=0, **options):
        # Check types
        try:
            double_page_element_type = corpus.types.get(slug=double_page_type)
        except ElementType.DoesNotExist:
            raise CommandError(f'Double page type {double_page_type} not found in {corpus}')

        try:
            single_page_element_type = corpus.types.get(slug=single_page_type)
        except ElementType.DoesNotExist:
            raise CommandError(f'Single page type {single_page_type} not found in {corpus}')

        # Retrieve all double pages that have children
        double_pages = corpus.elements.filter(type=double_page_element_type)
        if element:
            double_pages = double_pages.filter(id=element)
        total = double_pages.count()
        self.stdout.write(f"Found {total} double page(s) to process.")
        for i, double_page in enumerate(double_pages.iterator(1000)):
            try:
                self.stdout.write(f"Processing double page {double_page.id} ({i+1}/{total})")
                # Retrieve all children that are not a single page
                moved_children = corpus.elements.filter(paths__path__last=double_page.id).exclude(type=single_page_element_type).order_by('paths__ordering').distinct()
                if not moved_children.exists():
                    continue
                self.stdout.write(f"Found {len(moved_children)} child element that are not single pages.")

                # Retrieve all single pages
                single_pages = corpus.elements.filter(paths__path__last=double_page.id, type=single_page_element_type).distinct()
                self.stdout.write(f"Found {len(single_pages)} child single page(s).")

                for single_page in single_pages:
                    self.stdout.write(f"Processing single page {single_page.id}")
                    # Convert polygon to BBox to make computation easier
                    parent_polygon = reorder_polygon(single_page.polygon)

                    # find all children that are inside that page polygon (intersection over 50%)
                    close_enough_children = [children for children in moved_children if is_overlapping_enough(parent_polygon=parent_polygon, child_polygon=children.polygon, threshold=0.5)]
                    self.stdout.write(f"Found {len(close_enough_children)} child(ren) to move to this single page.")

                    # Update their path to make the single page their parent
                    for child in close_enough_children:
                        double_page.remove_child(child)
                        child.add_parent(single_page)
            except Exception as e:
                self.stdout.write(f"Processing has failed for double page ({double_page.id}) with error: {e}")
