from itertools import chain

import django
from django.db import DJANGO_VERSION_PICKLE_KEY, connections, models

from arkindex.project.fields import Unnest
from arkindex.users.models import Role


class ElementQuerySet(models.QuerySet):

    def __getstate__(self):
        """
        Django's QuerySets can be pickled, but when they are, they first call self._fetch_all()
        to fetch every row into the result cache and store it along with the pickled data,
        because one could expect to 'save' a queryset that way.
        We only pickle querysets for RQ tasks that will call .trash() later;
        this will never use the result cache as it does not load rows into memory, so we ignore that step.
        """
        return {**self.__dict__, DJANGO_VERSION_PICKLE_KEY: django.__version__}

    def trash(self, delete_children=True):
        """Performant deletion of any element queryset"""
        from arkindex.documents.models import (
            Classification,
            Element,
            MetaData,
            Selection,
            Transcription,
            TranscriptionEntity,
        )
        from arkindex.process.models import Process, ProcessElement, WorkerActivity

        # We only need the Element IDs to join on other models
        # Disable all ordering as it will not be required for deletions
        ids = self.order_by().values_list('id')

        # Generate the initial raw SQL query
        sql, params = ids.query.sql_with_params()

        # Delete all children from the targeted elements
        # using the same trash() method, iterating on every depth
        if delete_children is True:
            # First we need to find out the depths of the top elements' children
            # The fastest way is to calc all the lengths of children paths
            # by including the target SQL query, and joining it directly with paths
            # It's not possible to do that join with Django ORM
            with connections['default'].cursor() as cursor:
                cursor.execute("""select min(length), max(length) FROM (
                        select array_length(p.path, 1) as length
                        from documents_elementpath as p
                        inner join
                        ({}) as input on (array[input.id] && p.path)
                ) as lengths
                """.format(sql), params)
                min_paths, max_paths = cursor.fetchone()

            # Postgres will give us None when no children is found
            if min_paths is not None and max_paths is not None:
                # Django increments the path number to match the postgresl 1-index
                # Remove the lower levels first, then work the way up
                for i in reversed(range(min_paths - 1, max_paths)):
                    filters = {
                        f"paths__path__{i}__in": self,
                    }
                    Element.objects.filter(**filters).trash(delete_children=False)

        # TranscriptionEntity is linked through a transcription, so it's a bit harder to join
        TranscriptionEntity.objects.filter(transcription__element_id__in=ids).prefetch_related('transcription').delete()

        # We need to use raw deletion instead of the smarter delete from Django
        # to avoid loading in RAM the transcription ids to check that transcription entity is deleted cleanly
        Transcription.objects.filter(element_id__in=ids)._raw_delete(using='default')

        # Simple direct dependencies to remove, through a Foreign Key
        for cls in (WorkerActivity, MetaData, Classification, Selection, ProcessElement):
            cls.objects.filter(element_id__in=ids).delete()

        # Set elements directly on processes to None
        # This may cause workers processes to target a different set of elements
        Process.objects.filter(element_id__in=ids).update(element_id=None)

        # Set folder references on training processes to None
        Process.objects.filter(train_folder_id__in=ids).update(train_folder_id=None)
        Process.objects.filter(validation_folder_id__in=ids).update(validation_folder_id=None)
        Process.objects.filter(test_folder_id__in=ids).update(test_folder_id=None)

        # This is where it gets really ugly: in order to delete all the children
        # without losing their reference, we need to inject the generated SQL query
        # directly into an SQL DELETE statement for paths
        # Once paths are deleted, we can finally delete the targeted elements
        with connections['default'].cursor() as cursor:
            cursor.execute("""
            WITH element_ids (id) AS (
                DELETE FROM documents_elementpath
                WHERE element_id IN ({})
                RETURNING element_id
            )
            DELETE FROM documents_element element
            USING element_ids
            WHERE element.id = element_ids.id
            """.format(sql), params)

        # Finally, delete top elements.
        # Ensure the QuerySet does not use a DISTINCT; it is useless for a deletion, and since Django 3.2,
        # deleting causes a "TypeError: Cannot call delete() after .distinct()" error.
        # There is no QuerySet API to undo a distinct, so we edit the Query object directly
        self.query.distinct = False
        self.query.distinct_fields = ()
        self.only('id').delete()


class ElementManager(models.Manager):
    """Model manager for elements"""

    def get_queryset(self):
        return ElementQuerySet(self.model, using=self._db)

    def get_ascending(self, child_id, recursive=True):
        """
        Get all parent elements for a specific element ID.
        > All the elements from paths
        """
        from arkindex.documents.models import ElementPath

        paths = ElementPath.objects.filter(element_id=child_id)
        if recursive:
            parent_ids = paths.annotate(parent_id=Unnest('path')).values('parent_id')
        else:
            parent_ids = paths.values('path__last')

        return self.filter(id__in=parent_ids)

    def get_descending(self, parent_id, **filters):
        """
        Get all child elements for a specific element ID.
        > All the elements where parent is in the path
        """
        return self.filter(paths__path__contains=[parent_id], **filters).order_by('paths__ordering').distinct()

    def get_neighbor_paths(self, element):
        """
        Get previous and next element IDs for each path related to this element.

        This uses a single raw SQL query to use row numbers, instead of the ordering numeric values,
        to avoid issues with duplicated or missing orderings.

        Performance tests are available at https://gitlab.teklia.com/arkindex/backend/-/issues/1011

        * The elt_path subquery lists all paths for the current element.
        * The main query lists all neighbors for each path, retrieve the previous and next elements using
        a window function.
        * Paths are filtered depending on element corpus. That step requires another join, but allows top elements
        to be positioned correctly as their paths are similar among each corpus (i.e. empty).
        * The output is ordered by parent path, by ordering, then name and id for top elements.
        """
        from arkindex.documents.models import ElementPath
        return ElementPath.objects.raw(
            """
            with elt_path as (
              select path
                from documents_elementpath
                where element_id = %(element_id)s
            ), neighbors as (
              select
                neighbor.id,
                neighbor.path,
                neighbor.ordering,
                neighbor.element_id,
                lag(element_id) OVER (partition BY elt_path.path order by neighbor.ordering) as previous,
                lead(element_id) OVER (partition BY elt_path.path order by neighbor.ordering) as next,
                lag(ordering) OVER (partition BY elt_path.path order by neighbor.ordering) as previous_ord,
                lead(ordering) OVER (partition BY elt_path.path order by neighbor.ordering) as next_ord
              from
                documents_elementpath as neighbor
                inner join elt_path
                on neighbor.path = elt_path.path
                inner join documents_element
                on documents_element.id = neighbor.element_id
              where
                documents_element.corpus_id = %(element_corpus)s
              order by path, ordering, name, id
            )
            select *
              from neighbors
              where element_id = %(element_id)s
            """,
            params={
                'element_id': element.id,
                'element_corpus': element.corpus_id,
            },
        )

    def get_neighbors(self, element):
        """
        Returns a list of neighboring ElementPaths for an element, with a prefetched `element` attribute
        and a list of prefetched parent elements in the `parents` attribute.
        """
        paths = list(self.get_neighbor_paths(element))

        # Build a set of all IDs to load related elements (neighbors, parents) then load them into a dict
        related_elt_ids = set(chain(
            (element.id,),
            *((path.previous, path.next) for path in paths),
            *(path.path for path in paths),
        ))
        elements = {
            elt.id: elt
            for elt in (
                self.filter(id__in=filter(None, related_elt_ids))
                .select_related('type')
                .only('id', 'type__slug', 'name')
            )
        }

        # Generate an output corresponding to endpoint expectations (compatibility purpose)
        output = []
        for path in paths:
            if path.previous:
                output.append({
                    'ordering': path.previous_ord,
                    'element': elements.get(path.previous),
                    'parents': list(map(elements.get, path.path)),
                })
            output.append({
                'ordering': path.ordering,
                'element': elements.get(element.id),
                'parents': list(map(elements.get, path.path)),
            })
            if path.next:
                output.append({
                    'ordering': path.next_ord,
                    'element': elements.get(path.next),
                    'parents': list(map(elements.get, path.path)),
                })
        return output


class CorpusManager(models.Manager):
    '''
    Add ACL functions to corpus listing
    '''

    def filter_rights(self, *args, **kwargs):
        # Avoid circular dependencies as this module is imported by documents.models
        from arkindex.users.utils import filter_rights
        return filter_rights(*args, **kwargs)

    def readable(self, user):
        return super().get_queryset().filter(
            id__in=(self.filter_rights(user, self.model, Role.Guest.value).values('id'))
        )

    def writable(self, user):
        return super().get_queryset().filter(
            id__in=(self.filter_rights(user, self.model, Role.Contributor.value).values('id'))
        )

    def admin(self, user):
        return super().get_queryset().filter(
            id__in=(self.filter_rights(user, self.model, Role.Admin.value).values('id'))
        )
