from django.db import transaction
from django.db.models import F
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element, ElementPath
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import build_tree


class TestElementNeighbors(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug='volume')
        cls.page_type = cls.corpus.types.get(slug='page')

    def test_element_neighbors_acl(self):
        """
        A Guest access is required to list neighbors of an element
        """
        private_corpus = Corpus.objects.create(name="Private", public=False)
        private_type = private_corpus.types.create(slug="folder")
        elements = build_tree(
            {
                'B': 'A',
                'C': 'A',
                'D': 'A',
            },
            corpus=private_corpus,
            type=private_type,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:elements-neighbors', kwargs={'pk': str(elements['A'].id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a read access to this element.'})

    def test_element_neighbors(self):
        r"""
        A non authenticated user is able to list neighbors of a public element
            Z
            |
          X Y
         /|X \
        B A C D
        """
        elements = build_tree(
            {
                'Y': 'Z',
                'B': 'X',
                'A': ['X', 'Y'],
                'C': 'X',
                'D': 'Y',
            },
            corpus=self.corpus,
            type=self.volume_type,
        )

        response = self.client.get(reverse('api:elements-neighbors', kwargs={'pk': str(elements['A'].id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertListEqual(
            results,
            sorted([{
                # 2 neighbors via X (B, A, C) A has position 1
                'position': 0,
                'element': {
                    'id': str(elements['B'].id),
                    'type': 'volume',
                    'name': elements['B'].name
                },
                'parents': [{
                    'id': str(elements['X'].id),
                    'type': 'volume',
                    'name': elements['X'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['X'].id),
                    'type': 'volume',
                    'name': elements['X'].name
                }]
            }, {
                'position': 2,
                'element': {
                    'id': str(elements['C'].id),
                    'type': 'volume',
                    'name': elements['C'].name
                },
                'parents': [{
                    'id': str(elements['X'].id),
                    'type': 'volume',
                    'name': elements['X'].name
                }]
            }, {
                # 1 neighbors via Y (A, D) A has position 0
                'position': 0,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['Z'].id),
                    'type': 'volume',
                    'name': elements['Z'].name
                }, {
                    'id': str(elements['Y'].id),
                    'type': 'volume',
                    'name': elements['Y'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(elements['D'].id),
                    'type': 'volume',
                    'name': elements['D'].name
                },
                'parents': [{
                    'id': str(elements['Z'].id),
                    'type': 'volume',
                    'name': elements['Z'].name
                }, {
                    'id': str(elements['Y'].id),
                    'type': 'volume',
                    'name': elements['Y'].name
                }]
            }],
                # Order result by paths, as we can't predict the test generated uuids
                key=lambda path: ''.join([p['id'] for p in path['parents']]))
        )

    def test_element_neighbors_missing_ordering(self):
        """
        Check ElementNeighbors handles non-sequential orderings
        """
        elements = build_tree(
            {
                'A': 'W',
                'B': 'W',
                'C': 'W',
            },
            corpus=self.corpus,
            type=self.volume_type,
        )
        c_path = elements['C'].paths.get()
        c_path.ordering = 6584

        # The unique_element_orderings trigger will fail because an element is not supposed to have
        # multiple orderings within the same parent. This trigger is deferred, meaning it will only run
        # after the test ends, during teardown, so we cannot ignore it with pgtrigger.ignore().
        # Instead, we use a savepoint that we rollback immediately after the request, so that the trigger never runs.
        sid = transaction.savepoint()
        try:
            c_path.save()
            response = self.client.get(reverse('api:elements-neighbors', kwargs={'pk': str(elements['B'].id)}))
        finally:
            transaction.savepoint_rollback(sid)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertListEqual(
            results,
            [{
                'position': 0,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(elements['B'].id),
                    'type': 'volume',
                    'name': 'B'
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }, {
                'position': 6584,
                'element': {
                    'id': str(elements['C'].id),
                    'type': 'volume',
                    'name': elements['C'].name
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }]
        )

    def test_sub_neighbors(self):
        r"""
        Ensure the children of some elements of the same type do not interfere with the neighboring elements
           V
          /|\
         A B C
          / \
         B1 B2
        """
        elements = build_tree(
            {
                'A': 'V',
                'B': 'V',
                'C': 'V',
                'B1': 'B',
                'B2': 'B',
            },
            corpus=self.corpus,
            type=self.volume_type,
        )
        response = self.client.get(reverse('api:elements-neighbors', kwargs={'pk': str(elements['B'].id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']

        self.assertListEqual(
            results,
            [{
                'position': 0,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['V'].id),
                    'type': 'volume',
                    'name': elements['V'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(elements['B'].id),
                    'type': 'volume',
                    'name': 'B'
                },
                'parents': [{
                    'id': str(elements['V'].id),
                    'type': 'volume',
                    'name': elements['V'].name
                }]
            }, {
                'position': 2,
                'element': {
                    'id': str(elements['C'].id),
                    'type': 'volume',
                    'name': elements['C'].name
                },
                'parents': [{
                    'id': str(elements['V'].id),
                    'type': 'volume',
                    'name': elements['V'].name
                }]
            }]
        )

    def test_not_found(self):
        self.assertFalse(Element.objects.filter(id='aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa').exists())
        response = self.client.get(reverse(
            'api:elements-neighbors',
            kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'},
        ))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_no_duplicate_positions(self):
        """
        ListElementNeighbors had an issue where, if the element was found at position 10 in one parent
        and 20 in the other, elements at positions 9, 10, 11, 19, 20, and 21 were returned on both parents
        (12 rows instead of the expected 6).
             A            B
        ╭─┬─┬┴┬─┬─╮  ╭─┬─┬┴┬─┬─╮
        C X D E F G  H I J K X L
        """
        elements = build_tree(
            {
                'C': 'A',
                'D': 'A',
                'E': 'A',
                'F': 'A',
                'G': 'A',
                'H': 'B',
                'I': 'B',
                'J': 'B',
                'K': 'B',
                'L': 'B',
                'X': ['A', 'B'],
            },
            corpus=self.corpus,
            type=self.volume_type,
        )
        # The unique_element_orderings trigger will fail because an element is not supposed to have
        # multiple orderings within the same parent. This trigger is deferred, meaning it will only run
        # after the test ends, during teardown, so we cannot ignore it with pgtrigger.ignore().
        # Instead, we use a savepoint that we rollback immediately after the request, so that the trigger never runs.
        sid = transaction.savepoint()

        try:
            # Put X at the orderings we want it to have, then add 1 to D E F G and L to account for the move
            elements['X'].paths.filter(path__contains=[elements['A'].id]).update(ordering=1)
            elements['X'].paths.filter(path__contains=[elements['B'].id]).update(ordering=4)
            ElementPath.objects.filter(element__name__in=list('DEFGL')).update(ordering=F('ordering') + 1)

            # Ask for X's neighbors, should get C, X, D, K, X, and L.
            # Without the fix, would get the 12 children
            response = self.client.get(reverse('api:elements-neighbors', kwargs={'pk': str(elements['X'].id)}))
        finally:
            transaction.savepoint_rollback(sid)

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # The endpoint is ordered by ID, so depending on the random UUIDs we get, we can have CXDKXL or KXLCXD
        self.assertCountEqual(
            [result['element']['name'] for result in response.json()['results']],
            ['C', 'X', 'D', 'K', 'X', 'L'],
        )
