import shlex
import urllib.parse
import uuid
from os import path

from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.core.validators import MinLengthValidator
from django.db import models, transaction
from django.db.models import F, Q
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from enumfields import Enum, EnumField
from rest_framework.exceptions import ValidationError

import pgtrigger
from arkindex.documents.models import Classification, Element
from arkindex.images.models import ImageServer
from arkindex.ponos.models import FINAL_STATES, STATES_ORDERING, Artifact, State, Task, task_token_default
from arkindex.process.managers import ActivityManager, CorpusWorkerVersionManager, WorkerVersionManager
from arkindex.process.providers import get_provider
from arkindex.process.utils import get_default_farm_id
from arkindex.project.aws import S3FileMixin, S3FileStatus
from arkindex.project.fields import ArrayField, MD5HashField
from arkindex.project.models import IndexableModel
from arkindex.training.models import ModelVersionState


def get_task_depth(parents, slug):
    """
    Helper for Process.build_workflow that determines the depth of a task within a tree,
    without making any database queries, before the task is stored in the database.
    """
    parent_slugs = parents[slug]
    if not parent_slugs:
        return 0

    return max(
        get_task_depth(parents, parent_slug)
        for parent_slug in parent_slugs
    ) + 1


class ActivityState(Enum):
    """
    Store the state of the workers activity tracking for a process.
    To support large elements set, the state is asynchronously set to `ready` after a process
    has been started and worker activities have been initialized on its elements.
    """
    Disabled = 'disabled'
    Pending = 'pending'
    Ready = 'ready'
    Error = 'error'


class ProcessMode(Enum):
    Files = 'files'
    Repository = 'repository'
    IIIF = 'iiif'
    Workers = 'workers'
    Transkribus = 'transkribus'
    Template = 'template'
    S3 = 's3'
    Training = 'training'
    Local = 'local'
    Dataset = 'dataset'


class Process(IndexableModel):

    name = models.CharField(null=True, blank=True, max_length=100)
    creator = models.ForeignKey('users.User', on_delete=models.CASCADE, related_name='processes')
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='processes', null=True)
    mode = EnumField(ProcessMode, max_length=30)
    files = models.ManyToManyField('process.DataFile', related_name='processes')
    datasets = models.ManyToManyField('training.Dataset', related_name='processes', through='process.ProcessDataset')
    revision = models.ForeignKey(
        'process.Revision', related_name='processes', on_delete=models.CASCADE, blank=True, null=True)
    versions = models.ManyToManyField('process.WorkerVersion', through='process.WorkerRun', related_name='processes')
    activity_state = EnumField(ActivityState, max_length=32, default=ActivityState.Disabled)

    started = models.DateTimeField(blank=True, null=True)
    finished = models.DateTimeField(blank=True, null=True)

    farm = models.ForeignKey(
        to="ponos.Farm",
        related_name="processes",
        on_delete=models.DO_NOTHING,
        blank=True,
        null=True,
    )

    # Used to define the root element on which the process will run
    element = models.ForeignKey('documents.Element', on_delete=models.SET_NULL, null=True, blank=True)

    # Used to define the element type during folder elements creation for Images, PDF, IIIF and S3 imports
    folder_type = models.ForeignKey(
        'documents.ElementType',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='folder_imports',
    )

    # Used to define the element type during non-folder elements creation for Images, PDF, IIIF and S3 imports,
    # or to filter elements by type with Workers processes
    element_type = models.ForeignKey(
        'documents.ElementType',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='element_imports',
    )

    # Used to filter out elements with a name that doesn't contain the substring, only on Workers processes
    name_contains = models.CharField(null=True, blank=True, max_length=250)

    # Used to filter elements having a classification of a given MLClass in ListProcessElements
    ml_class = models.ForeignKey(
        'documents.MLClass',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='filtered_processes',
    )

    # Used to save a user's selection for Workers processes
    elements = models.ManyToManyField(
        'documents.Element',
        through='process.ProcessElement',
        related_name='processes',
    )

    # Used to include sub-elements of the actual query in Workers processes
    load_children = models.BooleanField(default=False)

    # Used to define the collection ID and entities import for Transkribus import
    collection_id = models.PositiveIntegerField(null=True, blank=True)

    # Use elements cache in Workers processes
    use_cache = models.BooleanField(default=False)

    # Allows to use a GPU in Workers processes
    use_gpu = models.BooleanField(default=False, blank=True)

    # Refers a process template
    template = models.ForeignKey(
        'process.Process',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='applications',
    )

    # S3 bucket name, and prefix within the bucket, for ProcessMode.S3
    # Bucket names are between 3 and 63 characters long
    bucket_name = models.CharField(
        max_length=63,
        blank=True,
        null=True,
        validators=[MinLengthValidator(3)],
    )
    # Object keys are up to 1024 characters long
    prefix = models.CharField(max_length=1024, blank=True, null=True)

    # Training specific fields
    model = models.ForeignKey(
        'training.Model',
        related_name='training_processes',
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    train_folder = models.ForeignKey(
        'documents.Element',
        related_name='train_folder_processes',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    validation_folder = models.ForeignKey(
        'documents.Element',
        related_name='validation_folder_processes',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    test_folder = models.ForeignKey(
        'documents.Element',
        related_name='test_folder_processes',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    class Meta:
        ordering = ['corpus', '-created']
        verbose_name_plural = 'processes'
        constraints = [
            models.CheckConstraint(
                # Either the process mode is local or repository and the process does not have a corpus,
                # or it has another mode and has a corpus set.
                check=Q(mode__in=(ProcessMode.Local, ProcessMode.Repository)) ^ ~Q(corpus=None),
                name='check_process_corpus',
                violation_error_message='Local and repository processes cannot have a corpus, and other modes must have one set.',
            ),
            models.UniqueConstraint(
                'creator',
                condition=Q(mode=ProcessMode.Local),
                name='unique_local_process',
                violation_error_message='Only one local process is allowed per user.',
            ),
            models.CheckConstraint(
                check=Q(started=None) | Q(started__gte=F('created')),
                name='process_started_after_created',
                violation_error_message='The process start date must not be earlier than the process creation date.',
            ),
            models.CheckConstraint(
                # This also implies that `started` must not be None if `finished` is set,
                # since finished >= NULL would be NULL and not TRUE.
                check=Q(finished=None) | Q(finished__gte=F('started')),
                name='process_finished_after_started',
                violation_error_message='The process finish date must not be earlier than the process start date.',
            ),
        ]

    @property
    def state(self) -> State:
        """
        Deduce the process state from the tasks of its latest run.
        A process' state is deduced by picking the first state that any tasks in this run have, in the order defined by STATES_ORDERING.
        """
        return self.get_state(self.get_last_run())

    @property
    def is_final(self) -> bool:
        return self.state in FINAL_STATES

    @property
    def has_prefetched_tasks(self):
        """
        Returns whether or not all tasks of this process have been prefetched with .prefetch_related()
        or prefetch_related_objects().
        Prefetched tasks will only be accessible through `self.tasks.all()` and using any other method will cause another SQL query.
        Therefore, if this property is True, only .all() should be used, and Python methods should be used to filter or aggregate, instead of `.exists`, `.filter` or `.aggregate`.

        See https://stackoverflow.com/a/19651840/5990435
        """
        return (
            hasattr(self, "_prefetched_objects_cache")
            and self.tasks.field.remote_field.get_cache_name()
            in self._prefetched_objects_cache
        )

    @property
    def expiry(self):
        """
        A process' expiry date. This is the latest expiry date of its tasks.
        No action is taken when a process is expired; the `cleanup` command alone uses it.
        """
        if self.has_prefetched_tasks:
            return max((t.expiry for t in self.tasks.all()), default=None)
        else:
            return self.tasks.aggregate(models.Max("expiry"))["expiry__max"]

    def get_last_run(self) -> int:
        """
        Get the last run number. If the `last_run` attribute is defined on this process,
        possibly from a `.annotate(last_run=Max('tasks__run'))` annotation in a Django QuerySet,
        this method will return the attribute's value instead of making another SQL query.
        """
        if not hasattr(self, "last_run"):
            # This prevents performing another SQL request when tasks have already been prefetched.
            # See https://stackoverflow.com/a/19651840/5990435
            if self.has_prefetched_tasks:
                self.last_run = max((t.run for t in self.tasks.all()), default=None)
            else:
                self.last_run = self.tasks.all().aggregate(models.Max("run"))["run__max"]

        if self.last_run is None:
            return -1

        return self.last_run

    def refresh_from_db(self, *args, **kwargs) -> None:
        # Remove the cached last_run if it was set, to ensure the process is properly refreshed
        if hasattr(self, 'last_run'):
            del self.last_run
        super().refresh_from_db(*args, **kwargs)

    def get_state(self, run) -> State:
        """
        A separate method to get a process' state on a given run.

        Most users will only use the Process.state property to get the state for a process' last run.
        However, when trying to get a state for many processes at once, using ``.annotate(last_run=Max('tasks__run'))``
        and using the annotation with this method will prevent many useless SQL requests.

        Further performance improvements can be achieved with ``prefetch_related('tasks')``.
        """
        # Negative run numbers never have tasks
        if run < 0:
            return State.Unscheduled

        # This prevents performing another SQL request when tasks have already been prefetched.
        # See https://stackoverflow.com/a/19651840/5990435
        if self.has_prefetched_tasks:
            task_states = set(t.state for t in self.tasks.all() if t.run == run)
        else:
            task_states = set(
                self.tasks.filter(run=run).values_list("state", flat=True)
            )

        # This run has no tasks
        if not task_states:
            return State.Unscheduled

        # All tasks have the same state
        if len(task_states) == 1:
            return task_states.pop()

        for state in STATES_ORDERING:
            if state in task_states:
                return state

        raise NotImplementedError("Something went wrong")

    def _get_filters(self):
        filters = {
            "corpus_id": self.corpus_id,
        }
        if self.name_contains:
            filters['name__icontains'] = self.name_contains
        if self.element_type_id:
            filters['type_id'] = self.element_type_id
        if self.ml_class_id:
            filters['id__in'] = Classification.objects.filter(ml_class_id=self.ml_class_id).values('element_id')

        return filters

    def copy_runs(self, new_process):
        """
        Copies this process' WorkerRuns to another process.

        This uses a dict to remap the existing parent UUIDs to the newly created UUIDs:
        otherwise, the runs on the other process would still have as their parents
        the runs of the current process, breaking their dependencies.
        """
        # Keep the relation between template_worker_runs and new_process_worker_runs ids
        new_runs = {}

        # First we copy each worker runs
        for run in self.worker_runs.all():
            # Create a new WorkerRun with same version, configuration and parents.
            new_run = WorkerRun(
                process=new_process,
                version_id=run.version_id,
                model_version_id=run.model_version_id,
                parents=run.parents,
                configuration_id=run.configuration_id,
                summary=run.summary
            )
            # Save the correspondence between this process' worker_run and the new one
            new_runs[run.id] = new_run

        # Remap parent ids correctly
        for run in new_runs.values():
            # If there are parents, we need to use the correct worker_run id, the newly created one
            run.parents = [new_runs[parent_id].id for parent_id in run.parents]
        WorkerRun.objects.bulk_create(new_runs.values())

    def list_elements(self):
        """
        Return a queryset of elements involved in this process
        """
        if self.mode != ProcessMode.Workers:
            return Element.objects.none()

        elements = None

        if self.element:
            # Assert the element has the same corpus as the process
            assert self.element.corpus_id == self.corpus_id, \
                f'Element {self.element.id} is not part of corpus {self.corpus_id}'
            # Handle a single element
            elements = Element.objects.filter(id=self.element.id)

        elif self.elements.exists():
            # Handle a selection of elements
            elements = self.elements.all()
            # Check all elements are in the same corpus as the process
            corpus = list(elements.values_list('corpus_id', flat=True).distinct())
            assert len(corpus) == 1 and corpus[0] == self.corpus_id, \
                f'Some elements on this process are not part of corpus {self.corpus_id}'

        if elements is not None and self.load_children:
            # Load all the children elements whose path contains the pre-selected elements
            # Those children are appended to the pre-selection
            elements |= Element.objects.filter(
                paths__path__overlap=map(str, elements.values_list('id', flat=True)),
            )

        # Load the full corpus, only when elements has not been populated before
        if elements is None:
            # Handle all elements of the process' corpus
            elements = Element.objects.filter(corpus=self.corpus_id)
            if not self.load_children:
                # Use top-level elements only
                elements = elements.filter(paths__path=[])

        # Filter elements depending on process properties
        return elements.filter(**self._get_filters())

    def build_task(
        self,
        command,
        slug,
        env,
        depth=0,
        run=0,
        image=None,
        parents=None,
        artifact=None,
        tags=None,
        has_docker_socket=False,
        extra_files=None,
        requires_gpu=False,
        shm_size=None
    ):
        if not image:
            image = settings.ARKINDEX_TASKS_IMAGE
        if not parents:
            parents = []
        if not tags:
            tags = []
        if not extra_files:
            extra_files = {}

        token = task_token_default()
        env['ARKINDEX_TASK_TOKEN'] = token

        task = Task(
            process=self,
            command=command,
            slug=slug,
            tags=tags,
            depth=depth,
            env=env,
            token=token,
            run=run,
            image=image,
            requires_gpu=requires_gpu,
            shm_size=shm_size,
            has_docker_socket=has_docker_socket,
            extra_files=extra_files,
            image_artifact_id=artifact,
        )
        # Avoid calling .set() when not necessary, as this can implicitly call task.save()
        if parents:
            task.parents.set(parents)
        return task

    @transaction.atomic
    def build_workflow(self, farm=None, chunks=None, thumbnails=False, corpus_id=None, **kwargs):
        """
        Create Ponos tasks according to configuration
        """
        ml_workflow_chunks = 1
        import_task = None
        import_task_name = 'import'
        # Use the default Ponos farm if no farm is specified
        self.farm_id = farm.id if farm is not None else get_default_farm_id()
        env = settings.PONOS_DEFAULT_ENV.copy()
        env['ARKINDEX_PROCESS_ID'] = str(self.pk)
        if self.corpus_id:
            env['ARKINDEX_CORPUS_ID'] = str(self.corpus_id)

        if self.mode == ProcessMode.Repository:
            assert self.revision is not None, \
                'A revision is required to create an import workflow from GitLab repository'
            if not self.revision.repo.enabled:
                raise ValidationError('Git repository does not have any valid credentials')
            import_task = self.build_task(
                command=(
                    f"python -m arkindex_tasks.import_git {self.revision.id}"
                    + (f' --corpus-id {corpus_id}' if corpus_id else '')
                ),
                slug=import_task_name,
                env=env
            )
            import_task.save()

        elif self.mode == ProcessMode.IIIF:
            worker_run = self.worker_runs.create(version=WorkerVersion.objects.imports_version, parents=[])
            env['ARKINDEX_WORKER_RUN_ID'] = str(worker_run.id)

            import_task = self.build_task(
                command=f'python -m arkindex_tasks.import_iiif.process {self.id}',
                slug=import_task_name,
                env=env
            )
            import_task.save()

        elif self.mode == ProcessMode.Transkribus:
            env['TRANSKRIBUS_EMAIL'] = settings.TRANSKRIBUS_EMAIL
            env['TRANSKRIBUS_PASSWORD'] = settings.TRANSKRIBUS_PASSWORD
            worker_run = self.worker_runs.create(version=WorkerVersion.objects.transkribus_version, parents=[])
            env['ARKINDEX_WORKER_RUN_ID'] = str(worker_run.id)

            import_task_name = 'import_arkindex'
            transkribus_task = self.build_task(
                command=f'python -m arkindex_tasks.export_transkribus {self.collection_id}',
                slug='export_transkribus',
                env=env
            )
            transkribus_task.save()
            import_task = self.build_task(
                command='python -m arkindex_tasks.import_transkribus '
                        f'--job-path /data/export_transkribus/transkribus_export_job.json '
                        f'--corpus {self.corpus.id}',
                slug=import_task_name,
                depth=1,
                parents=[transkribus_task],
                env=env
            )
            import_task.save()

        elif self.mode == ProcessMode.Files:
            worker_run = self.worker_runs.create(version=WorkerVersion.objects.imports_version, parents=[])
            env['ARKINDEX_WORKER_RUN_ID'] = str(worker_run.id)

            import_task = self.build_task(
                command=f'python -m arkindex_tasks.import_files {self.id}',
                slug=import_task_name,
                env=env
            )
            import_task.save()

        elif self.mode == ProcessMode.S3:
            assert self.bucket_name, 'Missing S3 bucket name'
            assert self.folder_type is not None, 'Missing folder type'
            assert self.element_type is not None, 'Missing element type'

            env['INGEST_S3_ACCESS_KEY'] = settings.INGEST_S3_ACCESS_KEY
            env['INGEST_S3_SECRET_KEY'] = settings.INGEST_S3_SECRET_KEY
            if settings.INGEST_S3_ENDPOINT:
                env['INGEST_S3_ENDPOINT'] = settings.INGEST_S3_ENDPOINT
            if settings.INGEST_S3_REGION:
                env['INGEST_S3_REGION'] = settings.INGEST_S3_REGION

            worker_run = self.worker_runs.create(version=WorkerVersion.objects.imports_version, parents=[])
            env['ARKINDEX_WORKER_RUN_ID'] = str(worker_run.id)

            command = (
                'python -m arkindex_tasks.import_s3'
                f' --corpus={self.corpus_id}'
                f' --bucket={shlex.quote(self.bucket_name)}'
                f' --folder-type={shlex.quote(self.folder_type.slug)}'
                f' --page-type={shlex.quote(self.element_type.slug)}'
                f' --iiif-base-url={shlex.quote(ImageServer.objects.ingest.url)}'
            )
            if self.prefix:
                command += f' --prefix={shlex.quote(self.prefix)}'
            if self.element:
                command += f' --element={self.element_id}'
            import_task = self.build_task(
                command=command,
                slug=import_task_name,
                env=env
            )
            import_task.save()

        elif self.mode == ProcessMode.Training:
            worker_version = kwargs.get('worker_version')
            model_version = kwargs.get('model_version')
            worker_configuration = kwargs.get('worker_configuration')
            assert worker_version, 'A worker version is required to start the training'

            worker_run = WorkerRun.objects.create(
                process_id=self.id,
                version_id=worker_version.id,
                configuration_id=worker_configuration and worker_configuration.id,
                parents=[],
                model_version=model_version,
            )

            env['ARKINDEX_WORKER_RUN_ID'] = str(worker_run.id)
            env['WORKER_VERSION_ID'] = str(worker_version.id)

            # Build the training task, as there is no initial task
            training_task = self.build_task(
                command=worker_version.docker_command,
                image=worker_version.docker_image_iid,
                slug='training',
                artifact=worker_version.docker_image_id,
                env=env,
                requires_gpu=self.use_gpu,
                shm_size=worker_version.docker_shm_size,
            )
            training_task.save()

        elif self.mode == ProcessMode.Workers:
            import_task_name = 'initialisation'
            if chunks is not None:
                assert chunks <= settings.MAX_CHUNKS, f'Import distribution is limited to {settings.MAX_CHUNKS} chunks'
                ml_workflow_chunks = chunks

            args = [
                'python', '-m', 'arkindex_tasks.init_elements', str(self.id),
                '--chunks-number', str(ml_workflow_chunks),
            ]
            if self.use_cache:
                args.append('--use-cache')

            import_task = self.build_task(
                command=' '.join(args),
                slug=import_task_name,
                env=env
            )
            import_task.save()

        elif self.mode == ProcessMode.Dataset:
            # For Dataset processes, there is no initial task
            import_task_name = None
            if chunks is not None:
                assert chunks <= settings.MAX_CHUNKS, f'Import distribution is limited to {settings.MAX_CHUNKS} chunks'
                ml_workflow_chunks = chunks

        else:
            raise NotImplementedError

        # Handle chunks and thumbnails generation on processes that list elements during the initial task
        # The S3 import does not generate a list of elements yet
        if self.mode in (ProcessMode.Workers, ProcessMode.Dataset) or (thumbnails and self.mode in (
            ProcessMode.Transkribus,
            ProcessMode.IIIF,
            ProcessMode.Files
        )):
            elts_chunk_files = ['elements.json']
            if ml_workflow_chunks > 1:
                elts_chunk_files = [f'elements_chunk_{n}.json' for n in range(1, ml_workflow_chunks + 1)]

            if self.mode in (ProcessMode.Workers, ProcessMode.Dataset):
                # Retrieve worker runs
                worker_runs = list(self.worker_runs.select_related('version__worker__repository', 'model_version').using('default'))
            else:
                worker_runs = []

            # Check that there are no GPU requirement conflicts before creating any task
            for worker_run in worker_runs:
                if not self.use_gpu and worker_run.version.gpu_usage == WorkerVersionGPUUsage.Required:
                    raise ValidationError(f"Worker version {worker_run.version.slug} requires a GPU and the `use_gpu` option is disabled.")

            tasks_to_create = []
            # Holds the task parents relationships, so that they can be built after building tasks
            # {task.slug: [parent_slug1, …]}
            if import_task:
                parents = {
                    import_task.slug: [],
                }
            else:
                parents = {}

            for index, elts_chunk in enumerate(elts_chunk_files, start=1):
                # Add a name suffix if task is distributed
                task_suffix = ''
                chunk = None
                if len(elts_chunk_files) > 1:
                    task_suffix = f'_{index}'
                    chunk = index

                elements_path = None
                if self.mode != ProcessMode.Dataset:
                    # No element handling in process for Dataset processes
                    elements_path = shlex.quote(path.join('/data', import_task_name, elts_chunk))

                # Generate thumbnails directly after import step
                if thumbnails:
                    # Thumbnails generation can't work with dataset processes because it requires the element.json that
                    # is generated in the initialisation task, and there is not initialisation task in dataset processes
                    assert self.mode != ProcessMode.Dataset, 'Thumbnails generation is incompatible with dataset mode processes.'

                    thumbnails_task = self.build_task(
                        command=f'python3 -m arkindex_tasks.generate_thumbnails {elements_path}',
                        slug=f'thumbnails{task_suffix}',
                        depth=1,
                        env=env,
                    )
                    tasks_to_create.append(thumbnails_task)
                    if import_task:
                        parents[thumbnails_task.slug] = [import_task.slug]
                    else:
                        parents[thumbnails_task.slug] = []

                # Generate a task for each WorkerRun on the Process
                for worker_run in worker_runs:
                    task_name = worker_run.version.slug
                    # The suffix is handled by WorkerRun.build_task
                    task, parent_slugs = worker_run.build_task(
                        self,
                        task_name,
                        env,
                        import_task_name,
                        elements_path,
                        chunk=chunk,
                        workflow_runs=worker_runs,
                    )
                    tasks_to_create.append(task)
                    parents[task.slug] = parent_slugs

            # Compute the depth of each task once the whole graph is built and before saving
            for task in tasks_to_create:
                task.depth = get_task_depth(parents, task.slug)

            # Build all tasks and task relationships in two queries
            Task.objects.bulk_create(tasks_to_create)

            # Use a tasks by slug map to find tasks more easily
            tasks = {task.slug: task for task in tasks_to_create}
            if import_task:
                tasks[import_task.slug] = import_task

            Task.parents.through.objects.bulk_create(
                Task.parents.through(
                    from_task=tasks[child_slug],
                    to_task=tasks[parent_slug],
                )
                for child_slug, parent_slugs in parents.items()
                for parent_slug in parent_slugs
            )

    def start(self, use_cache=False, worker_activity=False, use_gpu=False, **kwargs):
        if self.mode != ProcessMode.Workers:
            assert not use_cache and not worker_activity, 'Only worker processes can be started with cache or worker activities'

        if worker_activity:
            self.activity_state = ActivityState.Pending
        if use_cache:
            self.use_cache = True
        if use_gpu:
            self.use_gpu = True

        self.build_workflow(**kwargs)

        self.started = timezone.now()
        self.save()

        # Asynchronously initialize worker activities if required
        if self.activity_state == ActivityState.Pending:
            from arkindex.project.triggers import initialize_activity
            initialize_activity(self)

        if self.mode == ProcessMode.Workers:
            CorpusWorkerVersion.objects.bulk_create([
                CorpusWorkerVersion(
                    corpus_id=self.corpus_id,
                    worker_version_id=worker_version_id,
                    model_version_id=model_version_id,
                    worker_configuration_id=conf_id,
                )
                for worker_version_id, model_version_id, conf_id in self.worker_runs.values_list(
                    'version_id',
                    'model_version_id',
                    'configuration_id',
                )
            ], ignore_conflicts=True)

        worker_version_id = None
        if self.mode == ProcessMode.Transkribus:
            worker_version_id = settings.TRANSKRIBUS_WORKER_VERSION
        elif self.mode in (ProcessMode.Files, ProcessMode.IIIF, ProcessMode.S3):
            worker_version_id = settings.IMPORTS_WORKER_VERSION

        if worker_version_id:
            CorpusWorkerVersion.objects.get_or_create(corpus_id=self.corpus_id, worker_version_id=worker_version_id)

        # Save an SQL query when re-serializing this process: we know that the last run of this process is run 0 since we just started it.
        self.last_run = 0

    def retry(self):
        """
        Create new Task instances with a new run number and resets the completion date.

        :raises AssertionError: If the workflow is not in a final state.
        :returns: A dict mapping task slugs to task instances.
        :rtype: dict
        """
        # TODO: Maybe we should regenerate the tasks with build_workflow() instead of just cloning them like this.
        last_run = self.get_last_run()
        assert last_run < 0 or self.is_final, 'Cannot retry a process that is not in a final state.'

        # Retrying a process that never started is just the same as starting
        if last_run < 0:
            self.start()
            return

        if self.mode == ProcessMode.Repository and self.revision is not None and not self.revision.repo.enabled:
            raise ValidationError('Git repository does not have any valid credentials')

        new_tasks = {}
        task_parents = {}
        for task in self.tasks.all():
            new_task = Task(
                run=last_run + 1,
                depth=task.depth,
                slug=task.slug,
                priority=task.priority,
                tags=task.tags,
                image=task.image,
                shm_size=task.shm_size,
                command=task.command,
                env=task.env,
                has_docker_socket=task.has_docker_socket,
                image_artifact=task.image_artifact,
                agent_id=task.agent_id,
                requires_gpu=task.requires_gpu,
                gpu_id=task.gpu_id,
                process_id=task.process_id,
                container=task.container,
                extra_files=task.extra_files
            )

            # Set the task token on the new task
            new_task.env['ARKINDEX_TASK_TOKEN'] = new_task.token

            # TODO: Use a prefetch and task.parents.all() instead. This line causes 1 query per task
            task_parents[new_task.slug] = list(task.parents.values_list('slug', flat=True))
            new_tasks[task.slug] = new_task

        Task.objects.bulk_create(new_tasks.values())

        # Add parents once all the tasks are created
        for task in new_tasks.values():
            if task_parents[task.slug]:
                parents = [new_tasks[slug] for slug in task_parents[task.slug]]
                task.parents.set(parents)

        # setting last_run so that subsequent calls to get_last_run do not require db queries
        self.last_run = last_run + 1
        self.started = timezone.now()
        self.finished = None

        # If worker activities are enabled for this process, no matter their previous state,
        # re-initialize the activities just in case any were deleted or stolen by other processes
        # before we retried this one.
        if self.activity_state != ActivityState.Disabled:
            from arkindex.project.triggers import initialize_activity
            initialize_activity(self)
            self.activity_state = ActivityState.Pending

        self.save()

    def stop(self) -> None:
        """
        Fully stop the workflow by updating every running task to the Stopping state,
        and every unscheduled task to the Stopped state.
        """
        assert not self.is_final, 'Only processes in a final state can be stopped'

        stopping_count = self.tasks.filter(
            state__in=[State.Pending, State.Running]
        ).update(state=State.Stopping)
        self.tasks.filter(state=State.Unscheduled).update(state=State.Stopped)

        # If all the tasks are immediately stopped, then UpdateTask will not be able to update
        # the finished attribute, so we do it here.
        if not stopping_count:
            self.finished = timezone.now()
            self.save()

    def clear(self):
        self.worker_runs.all().delete()
        self.template_id = None
        self.save()


class ProcessElement(models.Model):
    """
    Link between Processes and Elements to allow workflows from selections.
    May be removed later once the ElementPath algorithms are optimized,
    as a selection could then be turned into a set.
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    process = models.ForeignKey(Process, on_delete=models.CASCADE, related_name='process_elements')
    element = models.ForeignKey('documents.Element', on_delete=models.CASCADE, related_name='process_elements')

    class Meta:
        unique_together = (
            ('process', 'element'),
        )


class ProcessDataset(models.Model):
    """
    Link between Processes and Datasets.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    process = models.ForeignKey(Process, on_delete=models.CASCADE, related_name='process_datasets')
    dataset = models.ForeignKey('training.Dataset', on_delete=models.CASCADE, related_name='process_datasets')

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['process', 'dataset'],
                name='unique_process_dataset',
            )
        ]


class DataFile(S3FileMixin, models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255)
    size = models.PositiveIntegerField(help_text='file size in bytes')
    content_type = models.CharField(max_length=120)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='files')
    status = EnumField(S3FileStatus, default=S3FileStatus.Unchecked, max_length=50)
    trashed = models.BooleanField(default=False)

    s3_bucket = settings.AWS_STAGING_BUCKET

    @cached_property
    def s3_key(self):
        return str(self.id)

    class Meta:
        ordering = ['corpus', 'name']

    def perform_check(self, save=True, raise_exc=False):
        """
        Check Datafile's existence, content type and size from S3
        """
        try:
            assert self.exists(), 'File has not been uploaded'
            assert self.s3_object.content_length == self.size, \
                f'Uploaded file size is {self.s3_object.content_length} bytes, expected {self.size} bytes'
            assert self.s3_object.content_type == self.content_type, \
                f'Uploaded file content type is {self.s3_object.content_type}, expected {self.content_type}'
            self.status = S3FileStatus.Checked
        except AssertionError:
            self.status = S3FileStatus.Error
            if raise_exc:
                raise
        finally:
            if save:
                self.save()


class Repository(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    url = models.URLField()
    hook_token = models.CharField(max_length=250)
    credentials = models.ForeignKey(
        'users.OAuthCredentials', on_delete=models.SET_NULL, related_name='repos', blank=True, null=True)
    git_ref_revisions = models.ManyToManyField('process.Revision', through='process.GitRef')
    memberships = GenericRelation('users.Right', 'content_id')

    class Meta:
        verbose_name_plural = 'repositories'
        constraints = [
            models.UniqueConstraint(
                'url',
                name='unique_repository_url',
            ),
            models.UniqueConstraint(
                'hook_token',
                name='unique_repository_hook_token',
            ),
        ]

    # Make this constant into a database field to be able to handle repositories from services other than GitLab
    provider_name = 'GitLabProvider'

    @property
    def provider_class(self):
        if not self.provider_name:
            raise ValueError("Empty provider_name")
        return get_provider(self.provider_name)

    @property
    def provider(self):
        if self.provider_class is None:
            raise NotImplementedError(f"Missing provider {self.provider_name}")
        return self.provider_class(credentials=self.credentials)

    @property
    def enabled(self):
        return self.credentials is not None and self.provider_class is not None

    def __str__(self):
        return f'{self.url}'


class GitRefType(Enum):
    Branch = 'branch'
    Tag = 'tag'


class GitRef(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    type = EnumField(GitRefType)
    name = models.CharField(max_length=250)
    repository = models.ForeignKey('process.Repository', on_delete=models.CASCADE, related_name='refs')
    revision = models.ForeignKey('process.Revision', on_delete=models.CASCADE, related_name='refs')

    class Meta:
        unique_together = (('name', 'repository'), )


class Revision(IndexableModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    repo = models.ForeignKey('process.Repository', on_delete=models.CASCADE, related_name='revisions')
    hash = models.CharField(max_length=50)
    message = models.TextField()
    author = models.CharField(max_length=50)

    class Meta:
        unique_together = (('repo', 'hash'), )

    @property
    def commit_url(self):
        return '{}/commit/{}'.format(self.repo.url.rstrip('/'), self.hash)

    def __str__(self):
        message = ' "{}"'.format(self.message.splitlines()[0]) if self.message else ''
        return '{}{} by {}'.format(self.hash[:8], message, self.author)


class Worker(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=100)
    slug = models.CharField(max_length=100)
    type = models.ForeignKey('process.WorkerType', on_delete=models.PROTECT, related_name='type')
    repository = models.ForeignKey('process.Repository', on_delete=models.CASCADE, related_name='workers')
    memberships = GenericRelation('users.Right', 'content_id')

    # Allow any authentified user to see and execute this worker
    public = models.BooleanField(default=False)

    class Meta:
        unique_together = (('slug', 'repository'),)

    def __str__(self):
        return f'{self.name}'


class WorkerType(IndexableModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = models.SlugField(max_length=100, db_index=False)
    display_name = models.CharField(max_length=100)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                'slug',
                name='unique_workertype_slug',
            ),
        ]

    def __str__(self):
        return self.slug

    def __repr__(self):
        return f'<WorkerType {self.slug}>'


class WorkerVersionState(Enum):
    Created = 'created'
    Processing = 'processing'
    Available = 'available'
    Error = 'error'


class WorkerVersionGPUUsage(Enum):
    Disabled = 'disabled'
    Supported = 'supported'
    Required = 'required'


class WorkerVersion(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    worker = models.ForeignKey('process.Worker', on_delete=models.CASCADE, related_name='versions')
    revision = models.ForeignKey('process.Revision', on_delete=models.CASCADE, related_name='versions')
    configuration = models.JSONField()
    state = EnumField(WorkerVersionState, default=WorkerVersionState.Created)
    gpu_usage = EnumField(WorkerVersionGPUUsage, default=WorkerVersionGPUUsage.Disabled, blank=True)
    model_usage = models.BooleanField(default=False)

    # Foreign key towards the ponos Artifact holding the image content
    docker_image = models.ForeignKey(Artifact, on_delete=models.SET_NULL, null=True, blank=True)

    # The Docker internal image id (sha256:xxx) that can be shared across multiple images
    docker_image_iid = models.CharField(null=True, blank=True, max_length=80)

    corpora = models.ManyToManyField(
        'documents.Corpus',
        through='process.CorpusWorkerVersion',
        related_name='worker_versions',
    )

    objects = WorkerVersionManager()

    class Meta:
        unique_together = (('worker', 'revision'),)
        constraints = [
            # Available worker versions must either have a docker_image_id or docker_image_iid set
            models.CheckConstraint(
                check=~models.Q(state=WorkerVersionState.Available) | ~Q(docker_image_id=None, docker_image_iid=None),
                name='workerversion_available_requires_docker_image',
            )
        ]

    def __str__(self):
        return f'{self.worker} for revision {self.revision}'

    @property
    def docker_image_name(self) -> str:
        parsed_url = urllib.parse.urlparse(self.worker.repository.url)
        return f'{parsed_url.netloc}{parsed_url.path}/{self.worker.slug}:{self.id}'.lower()

    @property
    def slug(self):
        return f'{self.worker.slug}_{self.truncated_id}'

    @property
    def truncated_id(self):
        return str(self.id)[0:6]

    @property
    def docker_command(self):
        if 'docker' not in self.configuration:
            return
        return self.configuration['docker'].get('command')

    @property
    def docker_shm_size(self):
        if 'docker' not in self.configuration:
            return
        return self.configuration['docker'].get('shm_size')


class WorkerConfiguration(IndexableModel):
    name = models.CharField(max_length=250)
    configuration = models.JSONField(default=dict)
    configuration_hash = MD5HashField()
    archived = models.BooleanField(default=False)
    worker = models.ForeignKey(
        Worker,
        on_delete=models.CASCADE,
        related_name='configurations',
    )

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'<WorkerConfiguration {self.name!r} ({self.id})>'

    class Meta:
        unique_together = (
            ('worker', 'configuration_hash'),
            ('worker', 'name')
        )
        constraints = [
            models.CheckConstraint(
                check=models.Q(configuration__typeof='object'),
                name='worker_configuration_configuration_objects',
            )
        ]


class WorkerRun(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    process = models.ForeignKey('process.Process', on_delete=models.CASCADE, related_name='worker_runs')
    version = models.ForeignKey('process.WorkerVersion', on_delete=models.CASCADE, related_name='worker_runs')
    model_version = models.ForeignKey('training.ModelVersion', on_delete=models.CASCADE, related_name='worker_runs', blank=True, null=True)
    parents = ArrayField(models.UUIDField(), blank=True, default=list)
    configuration = models.ForeignKey(
        WorkerConfiguration,
        related_name='worker_runs',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    summary = models.TextField()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['process', 'version'],
                condition=Q(model_version__isnull=True, configuration__isnull=True),
                name='unique_workerrun',
            ),
            models.UniqueConstraint(
                fields=['process', 'version', 'model_version'],
                condition=Q(model_version__isnull=False, configuration__isnull=True),
                name='unique_workerrun_model_version',
            ),
            models.UniqueConstraint(
                fields=['process', 'version', 'configuration'],
                condition=Q(model_version__isnull=True, configuration__isnull=False),
                name='unique_workerrun_configuration',
            ),
            models.UniqueConstraint(
                fields=['process', 'version', 'model_version', 'configuration'],
                condition=Q(model_version__isnull=False, configuration__isnull=False),
                name='unique_workerrun_model_version_and_configuration',
            ),
        ]

    def build_task(self, process, task_name, env, import_task_name, elements_path, run=0, chunk=None, workflow_runs=None):
        '''
        Build the Task that will represent this WorkerRun in ponos using :
        - the docker image name given by the WorkerVersion
        - the command given by the WorkerVersion
        - the artifact given by the WorkerVersion that will help us download the distant Docker image (S3)
        Return that Task and the list of its parent tasks slugs, to be added after they have all been created.
        '''
        task_env = env.copy()
        suffix = f'_{chunk}' if chunk else ''
        task_name = f'{task_name}{suffix}'

        token = task_token_default()
        task_env['ARKINDEX_TASK_TOKEN'] = token

        parents = None
        if self.parents:
            if workflow_runs:
                # Use already fetched worker runs if possible
                parent_runs = [run for run in workflow_runs if run.id in self.parents]
            else:
                parent_runs = (
                    WorkerRun
                    .objects
                    .using('default')
                    .filter(id__in=self.parents)
                    .select_related('version__worker')
                    # The IDs are required so that Django understands that everything is prefetched.
                    # WorkerVersion.slug will just require the version ID and worker slug
                    .only('id', 'version_id', 'version__worker_id', 'version__worker__slug')
                )
            parents = [
                f'{worker_run.version.slug}{suffix}'
                for worker_run in parent_runs
            ]
        elif import_task_name:
            parents = [import_task_name]
        else:
            parents = []

        assert (
            self.version.state == WorkerVersionState.Available
            and (self.version.docker_image_id or self.version.docker_image_iid)
        ), (
            f"Worker Version {self.version.id} is not available and cannot be used to build a task."
        )

        task_env["TASK_ELEMENTS"] = elements_path
        task_env["WORKER_VERSION_ID"] = str(self.version.id)
        task_env["ARKINDEX_WORKER_RUN_ID"] = str(self.id)
        if chunk:
            task_env['ARKINDEX_TASK_CHUNK'] = str(chunk)

        extra_files = {}
        # If the worker run needs a model_version, provide its public download url as extra file
        if self.model_version:
            assert self.model_version.state == ModelVersionState.Available, f'ModelVersion {self.model_version.id} is not available and cannot be used to build a task.'
            extra_files = {"model": settings.PUBLIC_HOSTNAME + reverse('api:model-version-download', kwargs={'pk': self.model_version.id}) + f"?token={self.model_version.build_authentication_token_hash()}"}

        task = Task(
            command=self.version.docker_command,
            image=self.version.docker_image_iid or self.version.docker_image_name,
            image_artifact_id=self.version.docker_image_id,
            env=task_env,
            shm_size=self.version.docker_shm_size,
            slug=task_name,
            depth=0,
            run=run,
            token=token,
            process=process,
            extra_files=extra_files,
            requires_gpu=process.use_gpu and self.version.gpu_usage in (WorkerVersionGPUUsage.Required, WorkerVersionGPUUsage.Supported)
        )

        return task, parents

    def build_summary(self):
        '''
        Build a summary of the worker run displaying information about :
        - worker version used
        - model version used if provided
        - worker configuration loaded if there is one
        '''
        summary_text = f"Worker {self.version.worker.name} @ "
        git_ref_names = self.version.revision.refs.values_list('name', flat=True)
        if len(git_ref_names) > 0:
            summary_text += ", ".join(git_ref_names)
        else:
            summary_text += self.version.truncated_id

        if self.model_version:
            summary_text += f" with model {self.model_version.model.name} @ {self.model_version.truncated_id}"

        if self.configuration:
            summary_text += f" using configuration '{self.configuration.name}'"

        self.summary = summary_text


class WorkerActivityState(Enum):
    Queued = 'queued'
    Started = 'started'
    Processed = 'processed'
    Error = 'error'


class WorkerActivity(models.Model):
    """
    Many-to-many relationship between Element and WorkerVersion
    Used to track the activity of a worker version among multiple elements
    """
    # Using an UUID helps to execute SQL raw INSERT
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    created = models.DateTimeField(auto_now_add=True)

    # We will manually handle the timestamp update using a PostgreSQL trigger to avoid issues with bulk inserts/updates,
    # and we still let Django handle setting the timestamp initially, so we use `auto_now_add` and not `auto_now`.
    updated = models.DateTimeField(auto_now_add=True)

    # To properly compute a processing time, we need to separately store the time at which the state was updated to `started`.
    started = models.DateTimeField(blank=True, null=True)

    element = models.ForeignKey(
        'documents.Element',
        on_delete=models.CASCADE,
        related_name='activities',
    )
    worker_version = models.ForeignKey(
        WorkerVersion,
        on_delete=models.CASCADE,
        related_name='activities',
    )
    state = EnumField(
        WorkerActivityState,
        default=WorkerActivityState.Queued,
    )
    process = models.ForeignKey(
        Process,
        related_name='activities',
        on_delete=models.CASCADE,
    )
    configuration = models.ForeignKey(
        WorkerConfiguration,
        related_name='worker_activities',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    # Specific WorkerActivity manager
    objects = ActivityManager()

    class Meta:
        constraints = [
            # Add configuration unicity
            # Two constraints are required as Null values are not compared for unicity
            models.UniqueConstraint(
                fields=['worker_version', 'element'],
                name='worker_activity_unique_no_configuration',
                condition=Q(configuration__isnull=True),
            ),
            models.UniqueConstraint(
                fields=['worker_version', 'element', 'configuration'],
                name='worker_activity_unique_configuration',
                condition=Q(configuration__isnull=False),
            ),
            # Require the `started` timestamp to be set if a WorkerActivity is in a `started` state.
            # This does not prevent activities in a processed/error state to not have start times,
            # as we do not know when activities that existed before the start time was added have started,
            # but is enough to ensure that we have to set the start time on any activity that now becomes started.
            models.CheckConstraint(
                check=~Q(state=WorkerActivityState.Started) | Q(started__isnull=False),
                name='worker_activity_started_requires_started',
            ),
        ]
        triggers = [
            pgtrigger.Trigger(
                name='update_workeractivity_updated',
                operation=pgtrigger.Update,
                when=pgtrigger.Before,
                func='NEW.updated = now(); RETURN NEW;',
            ),
            pgtrigger.ReadOnly(
                name='read_only_workeractivity_updated',
                fields=['updated'],
            )
        ]


class CorpusWorkerVersion(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    corpus = models.ForeignKey(
        'documents.Corpus',
        on_delete=models.CASCADE,
        related_name='worker_version_cache',
    )
    worker_version = models.ForeignKey(
        WorkerVersion,
        on_delete=models.CASCADE,
        related_name='corpus_cache',
    )
    model_version = models.ForeignKey(
        'training.ModelVersion',
        related_name='corpus_worker_versions',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    worker_configuration = models.ForeignKey(
        WorkerConfiguration,
        related_name='corpus_worker_versions',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    objects = CorpusWorkerVersionManager()

    class Meta:
        constraints = [
            # A worker version can be linked to a corpus via multiple model version or configuration
            models.UniqueConstraint(
                fields=['corpus', 'worker_version'],
                condition=Q(model_version__isnull=True, worker_configuration__isnull=True),
                name='corpus_workerversion_version_null_configuration_null',
            ),
            models.UniqueConstraint(
                fields=['corpus', 'worker_version', 'model_version'],
                condition=Q(model_version__isnull=False, worker_configuration__isnull=True),
                name='corpus_workerversion_version_not_null_configuration_null',
            ),
            models.UniqueConstraint(
                fields=['corpus', 'worker_version', 'worker_configuration'],
                condition=Q(model_version__isnull=True, worker_configuration__isnull=False),
                name='corpus_workerversion_version_null_configuration_not_null',
            ),
            models.UniqueConstraint(
                fields=['corpus', 'worker_version', 'model_version', 'worker_configuration'],
                condition=Q(model_version__isnull=False, worker_configuration__isnull=False),
                name='corpus_workerversion_version_not_null_configuration_not_null',
            ),
        ]
