from unittest.mock import patch

from django.urls import reverse
from django.utils import timezone as DjangoTimeZone
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.process.models import Process, ProcessMode
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import fake_now
from arkindex.training.models import Dataset, DatasetState
from arkindex.users.models import User

# Using the fake DB fixtures creation date when needed
FAKE_CREATED = '2020-02-02T01:23:45.678000Z'


class TestDatasetsAPI(FixtureAPITestCase):
    @classmethod
    @patch.object(DjangoTimeZone, 'now', fake_now)
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = Process.objects.get(mode=ProcessMode.Workers)
        cls.process.start()
        cls.task = cls.process.tasks.first()
        cls.dataset_creator = User.objects.create(email='adam@nerv.co.jp', display_name='Adam', verified_email=True)
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.read_user = User.objects.get(email='user3@user.fr')
        cls.write_user = User.objects.get(email='user2@user.fr')
        cls.dataset = Dataset.objects.get(name='First Dataset')
        cls.dataset2 = Dataset.objects.get(name='Second Dataset')
        cls.private_dataset = Dataset.objects.create(name="Private Dataset", description="Dead Sea Scrolls", corpus=cls.private_corpus, creator=cls.dataset_creator)
        cls.dataset.task = cls.task
        cls.dataset.save()

    # ListCorpusDatasets

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_list_requires_verified(self):
        user = User.objects.create(email='not_verified@mail.com', display_name='Not Verified', verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_list_corpus_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:corpus-datasets', kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_list_private_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:corpus-datasets', kwargs={'pk': self.private_corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have guest access to this corpus."})

    def test_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['results'], [
            {
                'name': 'First Dataset',
                'description': 'dataset number one',
                'id': str(self.dataset.pk),
                'sets': ['training', 'test', 'validation'],
                'state': 'open'
            },
            {
                'name': 'Second Dataset',
                'description': 'dataset number two',
                'id': str(self.dataset2.pk),
                'sets': ['training', 'test', 'validation'],
                'state': 'open'
            }
        ])

    # CreateDataset

    def test_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset', 'description': 'My dataset for my experiments.'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_create_requires_verified(self):
        user = User.objects.create(email='not_verified@mail.com', display_name='Not Verified', verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset', 'description': 'My dataset for my experiments.'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_create_corpus_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'}),
                data={'name': 'My dataset', 'description': 'My dataset for my experiments.'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_create_private_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.private_corpus.pk}),
                data={'name': 'My dataset', 'description': 'My dataset for my experiments.'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to this corpus."})

    def test_create_name_required(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'description': 'My dataset for my experiments.'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['This field is required.']})

    def test_create_name_empty(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': '', 'description': 'My dataset for my experiments.'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['This field may not be blank.']})

    def test_create_name_blank(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': '        ', 'description': 'My dataset for my experiments.'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['This field may not be blank.']})

    def test_create_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={
                    'name': 'In elementum leo et lectus finibus, eget lobortis ligula posuere. Vivamus finibus quis sapien in mattis. Cras in libero.',
                    'description': 'My dataset for my experiments.'
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['Ensure this field has no more than 100 characters.']})

    def test_create_name_already_exists_in_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={
                    'name': 'First Dataset',
                    'description': 'My dataset for my experiments.'
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'non_field_errors': ['A dataset with this name already exists in this corpus.']})

    def test_create_description_required(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'description': ['This field is required.']})

    def test_create_description_empty(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset', 'description': ''},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'description': ['This field may not be blank.']})

    def test_create_description_blank(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset', 'description': '             '},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'description': ['This field may not be blank.']})

    def test_create(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset', 'description': 'My dataset for my experiments.'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_dataset = Dataset.objects.get(name='My dataset')
        self.assertDictEqual(response.json(), {
            "id": str(created_dataset.id),
            "name": "My dataset",
            "description": "My dataset for my experiments.",
            "sets": ['training', 'test', 'validation'],
            "state": "open"
        })

    def test_create_state_ignored(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset', 'description': 'My dataset for my experiments.', 'state': 'complete'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_dataset = Dataset.objects.get(name='My dataset')
        self.assertEqual(created_dataset.state, DatasetState.Open)
        self.assertDictEqual(response.json(), {
            "id": str(created_dataset.id),
            "name": "My dataset",
            "description": "My dataset for my experiments.",
            "sets": ['training', 'test', 'validation'],
            "state": "open"
        })

    def test_create_sets(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset', 'description': 'My dataset for my experiments.', 'sets': ['a', 'b', 'c', 'd']},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_dataset = Dataset.objects.get(name='My dataset')
        self.assertDictEqual(response.json(), {
            "id": str(created_dataset.id),
            "name": "My dataset",
            "description": "My dataset for my experiments.",
            "sets": ['a', 'b', 'c', 'd'],
            "state": "open"
        })

    def test_create_sets_length(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset', 'description': 'My dataset for my experiments.', 'sets': []},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'sets': ['Either do not specify set names to use the default values, or specify a non-empty list of names.']})

    def test_create_sets_unique_names(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset', 'description': 'My dataset for my experiments.', 'sets': ['a', 'a', 'b']},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'sets': ['Set names must be unique.']})

    def test_create_sets_blank_names(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={'name': 'My dataset', 'description': 'My dataset for my experiments.', 'sets': ['     ', ' ', 'b']},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'sets':
                {
                    '0': ['This field may not be blank.'],
                    '1': ['This field may not be blank.']
                }
        })

    def test_create_sets_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:corpus-datasets', kwargs={'pk': self.corpus.pk}),
                data={
                    'name': 'My Dataset',
                    'description': 'My dataset for my experiments.',
                    'sets': ['unit-00', 'Etiam accumsan ullamcorper mauris eget mattis. Ut porttitor.']
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'sets': {
                '1': ['Ensure this field has no more than 50 characters.']
            }
        })

    # UpdateDataset

    def test_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'Shin Seiki Evangelion',
                    'description': 'Omedeto!',
                    'sets': ['unit-01', 'unit-00', 'unit-02'],
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_update_requires_verified(self):
        user = User.objects.create(email='not_verified@mail.com', display_name='Not Verified', verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'Shin Seiki Evangelion',
                    'description': 'Omedeto!',
                    'sets': ['unit-01', 'unit-00', 'unit-02'],
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_update_requires_write_corpus(self):
        self.client.force_login(self.read_user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'Shin Seiki Evangelion',
                    'description': 'Omedeto!',
                    'sets': ['unit-01', 'unit-00', 'unit-02'],
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to corpus Unit Tests."})

    def test_update_doesnt_exist(self):
        self.client.force_login(self.read_user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'}),
                data={
                    'name': 'Shin Seiki Evangelion',
                    'description': 'Omedeto!',
                    'sets': ['unit-01', 'unit-00', 'unit-02'],
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_update_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'In elementum leo et lectus finibus, eget lobortis ligula posuere. Vivamus finibus quis sapien in mattis. Cras in libero.',
                    'description': 'My dataset for my experiments.',
                    'sets': ['unit-01', 'unit-00', 'unit-02'],
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['Ensure this field has no more than 100 characters.']})

    def test_update_name_already_exists_in_corpus(self):
        Dataset.objects.create(name="Another Dataset", description="A set of data", corpus=self.corpus, creator=self.dataset_creator)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'Another Dataset',
                    'description': 'My dataset for my experiments.',
                    'sets': ['unit-01', 'unit-00', 'unit-02'],
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'non_field_errors': ['A dataset with this name already exists in this corpus.']})

    def test_update_requires_all_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'Shin Seiki Evangelion',
                    'sets': ['unit-01', 'unit-00', 'unit-02'],
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"description": ["This field is required."]})

    def test_update(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'Shin Seiki Evangelion',
                    'description': 'Omedeto!',
                    'sets': ['unit-01', 'unit-00', 'unit-02']
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)
        self.assertEqual(self.dataset.name, 'Shin Seiki Evangelion')
        self.assertEqual(self.dataset.description, 'Omedeto!')
        self.assertListEqual(self.dataset.sets, ['unit-01', 'unit-00', 'unit-02'])

    def test_update_sets_length(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'Shin Seiki Evangelion',
                    'description': 'Omedeto!',
                    'sets': []
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'sets': ['Either do not specify set names to use the default values, or specify a non-empty list of names.']})

    def test_update_sets_unique_names(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'Shin Seiki Evangelion',
                    'description': 'Omedeto!',
                    'sets': ['unit-01', 'unit-01', 'unit-02']
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'sets': ['Set names must be unique.']})

    def test_update_sets_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'My Dataset',
                    'description': 'My dataset for my experiments.',
                    'sets': ['unit-00', 'Etiam accumsan ullamcorper mauris eget mattis. Ut porttitor.']
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'sets': {
                '1': ['Ensure this field has no more than 50 characters.']
            }
        })

    def test_update_empty_or_blank_description_or_name(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': '',
                    'description': '          ',
                    'sets': ['unit-00', 'unit-01', 'unit-02']
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['This field may not be blank.'], 'description': ['This field may not be blank.']})

    def test_update_all_errors(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': '',
                    'description': '          ',
                    'sets': ['unit-00', 'unit-00', 'unit-02']
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'name': ['This field may not be blank.'],
            'description': ['This field may not be blank.'],
            'sets': ['Set names must be unique.']
        })

    def test_update_not_ponos_state_ignored(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'Shin Seiki Evangelion',
                    'description': 'Omedeto!',
                    'sets': ['unit-01', 'unit-00', 'unit-02'],
                    'state': 'complete'
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)

    def test_update_ponos_task_state_update(self):
        with self.assertNumQueries(7):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    'name': 'Shin Seiki Evangelion',
                    'description': 'Omedeto!',
                    'sets': ['unit-01', 'unit-00', 'unit-02'],
                    'state': 'complete'
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Complete)

    def test_update_ponos_task_bad_state(self):
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    'name': 'Shin Seiki Evangelion',
                    'description': 'Omedeto!',
                    'sets': ['unit-01', 'unit-00', 'unit-02'],
                    'state': 'decommissioned'
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'state': ['Value is not of type DatasetState']})
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)

    # PartialUpdateDataset

    def test_partial_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={'name': 'Shin Seiki Evangelion'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_partial_update_requires_verified(self):
        user = User.objects.create(email='not_verified@mail.com', display_name='Not Verified', verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={'name': 'Shin Seiki Evangelion'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_partial_update_requires_write_corpus(self):
        self.client.force_login(self.read_user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={'name': 'Shin Seiki Evangelion'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to corpus Unit Tests."})

    def test_partial_update_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'In elementum leo et lectus finibus, eget lobortis ligula posuere. Vivamus finibus quis sapien in mattis. Cras in libero.',
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['Ensure this field has no more than 100 characters.']})

    def test_partial_update_name_already_exists_in_corpus(self):
        Dataset.objects.create(name="Another Dataset", description="A set of data", corpus=self.corpus, creator=self.dataset_creator)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': 'Another Dataset',
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'non_field_errors': ['A dataset with this name already exists in this corpus.']})

    def test_partial_update(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'description': 'Omedeto!',
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)
        self.assertEqual(self.dataset.name, 'First Dataset')
        self.assertEqual(self.dataset.description, 'Omedeto!')
        self.assertListEqual(self.dataset.sets, ['training', 'test', 'validation'])

    def test_partial_update_empty_or_blank_description_or_name(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'name': '                  ',
                    'description': '',
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['This field may not be blank.'], 'description': ['This field may not be blank.']})

    def test_partial_update_not_ponos_state_ignored(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'state': 'complete'
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)

    def test_partial_update_ponos_task_state_update(self):
        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    'state': 'complete'
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Complete)

    def test_partial_update_ponos_task_bad_state(self):
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    'state': 'decommissioned'
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'state': ['Value is not of type DatasetState']})
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)

    def test_partial_update_sets_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                data={
                    'sets': ['unit-00', 'Etiam accumsan ullamcorper mauris eget mattis. Ut porttitor.']
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'sets': {
                '1': ['Ensure this field has no more than 50 characters.']
            }
        })

    # RetrieveDataset

    def test_retrieve_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_retrieve_requires_verified(self):
        user = User.objects.create(email='not_verified@mail.com', display_name='Not Verified', verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_retrieve_requires_read_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:dataset-update', kwargs={'pk': self.private_dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_retrieve_doesnt_exists(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:dataset-update', kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_retrieve(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.dataset.id),
            "name": "First Dataset",
            "description": "dataset number one",
            "state": "open",
            "sets": ["training", "test", "validation"],
            "creator": "Test user",
            "task_id": str(self.task.id),
            "corpus_id": str(self.corpus.id),
            "created": FAKE_CREATED,
            "updated": FAKE_CREATED
        })

    # DestroyDataset

    def test_delete_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_delete_requires_verified(self):
        user = User.objects.create(email='not_verified@mail.com', display_name='Not Verified', verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.delete(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_delete_requires_write_corpus(self):
        self.client.force_login(self.write_user)
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to corpus Unit Tests."})

    def test_delete_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.delete(
                reverse('api:dataset-update', kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'}),
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_delete(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.delete(
                reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Dataset.DoesNotExist):
            self.dataset.refresh_from_db()

    def test_delete_not_open(self):
        self.client.force_login(self.user)
        cases = [DatasetState.Building, DatasetState.Complete, DatasetState.Error]
        for state in cases:
            with self.subTest(state=state):
                self.dataset.state = state
                self.dataset.save()
                with self.assertNumQueries(5):
                    response = self.client.delete(
                        reverse('api:dataset-update', kwargs={'pk': self.dataset.pk}),
                    )
                    self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
                self.assertDictEqual(response.json(), {"detail": "Only datasets in an open state can be deleted."})
